import sys
import wx
from wx import xrc

from matplotlib.figure import Figure
from matplotlib.backends.backend_wxagg import \
        FigureCanvasWxAgg as FigCanvas, \
        NavigationToolbar2WxAgg as NavigationToolbar
from plot_utils import ImagePlot
from square_selector import SquareSelector
from reflexlib import VisirParameter
from reflex import RecipeParameter
import numpy as np
import os
from wx.lib.mixins.listctrl import CheckListCtrlMixin


def add_dialog_buttons(parent, on_help=None):
    btnSizer = wx.StdDialogButtonSizer()
    btnSizerOK = wx.Button(parent, wx.ID_OK)
    btnSizer.AddButton(btnSizerOK)
    btnSizerCancel = wx.Button(parent, wx.ID_CANCEL)
    btnSizer.AddButton(btnSizerCancel)
    if on_help:
        btnSizerHelp = wx.Button(parent, wx.ID_HELP)
        btnSizer.AddButton(btnSizerHelp)
        btnSizerHelp.Bind(wx.EVT_BUTTON, on_help)
    btnSizer.Realize()
    return btnSizer


class PlotPanel(wx.Panel):
    def __init__(self, parent, toolbar=True, id=wx.ID_ANY, **kwargs):
        wx.Panel.__init__(self, parent, id=id, **kwargs)
        self.figure = Figure(figsize=(2, 2))
        self.canvas = FigCanvas(self, -1, self.figure)

        sizer = wx.BoxSizer(wx.VERTICAL)
        if toolbar:
            self.toolbar = NavigationToolbar(self.canvas)
            self.toolbar.Realize()
            sizer.Add(self.toolbar, 0, wx.LEFT | wx.EXPAND)
        sizer.Add(self.canvas, 1, wx.EXPAND)
        self.SetSizer(sizer)


class PlotDialog(wx.Dialog):
    def __init__(self, parent, *args, **kwargs):
        if 'size' not in kwargs:
            kwargs['size'] = wx.Size(500, 500)
        if 'style' not in kwargs:
            kwargs['style'] = (wx.CLOSE_BOX | wx.DEFAULT_DIALOG_STYLE |
                               wx.MAXIMIZE_BOX | wx.MINIMIZE_BOX |
                               wx.RESIZE_BORDER)
        wx.Dialog.__init__(self, parent, *args, **kwargs)
        sizer = wx.BoxSizer(wx.VERTICAL)

        plotPanel = PlotPanel(self)
        self.figure = plotPanel.figure

        sizer.Add(plotPanel, 1, wx.EXPAND)
        sizer.Add(add_dialog_buttons(self), 0, wx.EXPAND)
        sizer.AddSpacer((0, 3), 0, wx.EXPAND, 5)
        self.SetSizer(sizer)


class CheckListCtrl(wx.ListCtrl, CheckListCtrlMixin):
    def __init__(self, parent, id, pos, size):
        wx.ListCtrl.__init__(self, parent, id, pos, size, style=wx.LC_REPORT)
        CheckListCtrlMixin.__init__(self)


class BeamDialog(wx.Dialog):
    def __init__(self, parent, data=None, dflt_size=50, **kwargs):
        if 'size' not in kwargs:
            kwargs['size'] = wx.Size(500, 500)
        if 'style' not in kwargs:
            kwargs['style'] = (wx.CLOSE_BOX | wx.DEFAULT_DIALOG_STYLE |
                               wx.MAXIMIZE_BOX | wx.MINIMIZE_BOX |
                               wx.RESIZE_BORDER)
        wx.Dialog.__init__(self, parent, **kwargs)
        sizer = wx.BoxSizer(wx.VERTICAL)

        mainsizer = wx.FlexGridSizer(0, 2, 0, 0)
        mainsizer.AddGrowableCol(0)
        mainsizer.AddGrowableRow(0)
        mainsizer.SetFlexibleDirection(wx.BOTH)
        mainsizer.SetNonFlexibleGrowMode(wx.FLEX_GROWMODE_SPECIFIED)

        ctrlSizer = wx.BoxSizer(wx.VERTICAL)
        self.ctrlPanel = wx.Panel(self, wx.ID_ANY, wx.DefaultPosition,
                                  wx.Size(-1, -1), wx.TAB_TRAVERSAL)

        self.beamBox = CheckListCtrl(self.ctrlPanel, wx.ID_ANY,
                                     wx.DefaultPosition,
                                     wx.Size(150, 150))
        self.beamBox.InsertColumn(0, 'Sign', width=35)
        self.beamBox.InsertColumn(1, 'Position', width=115)
        ctrlSizer.Add(self.beamBox, 0, wx.ALL | wx.EXPAND, 5)

        self.ctrlPanel.SetSizer(ctrlSizer)
        self.ctrlPanel.Layout()

        plotPanel = PlotPanel(self)
        self.figure = plotPanel.figure

        mainsizer.Add(plotPanel, 1, wx.EXPAND)
        mainsizer.Add(self.ctrlPanel, 1, wx.EXPAND | wx.ALL, 5)

        sizer.Add(mainsizer, 1, wx.EXPAND)
        sizer.Add(add_dialog_buttons(self, on_help=self.on_help), 0, wx.EXPAND)
        sizer.AddSpacer((0, 3), 0, wx.EXPAND, 5)
        self.SetSizer(sizer)

        self.data = data
        plt = ImagePlot(self.figure, self.figure.add_subplot(111))
        plt.set_data(data)
        plt.ax.set_title('Input image')
        plt.draw(origin='lower', interpolation='nearest')
        self.sq = SquareSelector(plt.ax, self.on_select, dflt_size=dflt_size)

    def on_help(self, event):
        msg = """
Select the position of the beams in the image with the left mouse button.
The selection does not need to be perfect. The values are only used an initial guess and are refined by the recipe later.

By holding the mouse button and dragging you can change the size of the image around the beam.
Selections can be removed by right clicking them.

The pipeline also needs to know the sign of the beam as due to the chop/nod correction some objects are negative and some positive.
It will try to detect the sign from the selected region.
If the autodetection fails it will prompt for the sign.
The detected or chosen sign can also be changed by clicking the checkbox next to the beam entry on the right.
A checked box represents a positive beam and an unchecked box a negative beam.
"""
        dlg = wx.MessageDialog(self, msg,
                               "Help",
                               wx.OK | wx.ICON_INFORMATION)
        dlg.ShowModal()
        dlg.Destroy()

    def search_bbox(self, v, col):
        for idx in range(self.beamBox.GetItemCount()):
            item = self.beamBox.GetItem(idx, col)
            if item.GetText() == v:
                    return idx

    def check_sign(self, x, y, w, d):
        lx = int(max(0, x - w / 3))
        ly = int(max(0, y - w / 3))
        hx = int(min(d.shape[1], x + w / 3))
        hy = int(min(d.shape[0], y + w / 3))
        dc = d[ly:hy, lx:hx] - np.median(d[ly:hy, lx:hx])
        if abs(dc.sum()) < dc.std() * np.sqrt(dc.size) * 2:
            while True:
                sdlg = wx.TextEntryDialog(self, 'Beam (%d,%d) sign pos or neg?'
                                          % (int(x), int(y)),
                                         'Set beam sign')
                sdlg.ShowModal()
                sign = sdlg.GetValue().lower()
                sdlg.Destroy()
                if sign in ['pos', 'neg']:
                    break
        else:
            sign = (dc).sum()
            sign = 'pos' if sign > 0 else 'neg'
        return sign

    def on_select(self, data, add):
        st = '%d, %d, %d' % (int(data[0]), int(data[1]), int(data[2]) * 2)
        ind = self.search_bbox(st, 1)
        if not add and ind >= 0:
            self.beamBox.DeleteItem(ind)
        else:
            idx = self.beamBox.InsertStringItem(sys.maxint, '')
            sign = self.check_sign(*data, d=self.data)
            if sign == 'pos':
                self.beamBox.CheckItem(idx)
            self.beamBox.SetStringItem(idx, 1, st)

    def get_selections(self):
        res = []
        for idx in range(self.beamBox.GetItemCount()):
            sign = 'pos' if self.beamBox.IsChecked(idx)else 'neg'
            pos = map(int, self.beamBox.GetItem(idx, 1).GetText().split(','))
            res.append((sign, pos))
        return res


class ParamWidget:
    def __init__(self, parent, param):
        self.param = param
        self.parent = parent
        self.paramLabel = wx.StaticText(parent, label=param.display_name)
        if param.type == 'bool':
            self.paramCtrl = wx.CheckBox(parent, wx.ID_ANY, u"")
        elif param.type == 'list':
            self.paramCtrl = \
                wx.ComboBox(parent, wx.ID_ANY, value=param.value,
                            style=wx.CB_READONLY, choices=param.options)
        elif param.type == 'file':
            self.paramCtrl = wx.FilePickerCtrl(parent, wx.ID_ANY,
                                               wildcard='*.fits')
        else:
            self.paramCtrl = wx.TextCtrl(parent, wx.ID_ANY,
                                         style=(wx.TE_PROCESS_ENTER |
                                                wx.TE_RIGHT))
        self.set_value(param.value)
        self.default = self.get_value()
        if param.description:
            self.paramCtrl.SetToolTipString(param.description)

    def get_value(self):
        try:
            v = self.paramCtrl.GetValue()
        except AttributeError:
            v = self.paramCtrl.GetPath()
        return self.param.to_string(self.param.type, v)

    def is_default(self):
        return self.get_value() == self.default

    def set_value(self, value):
        try:
            self.paramCtrl.SetValue(value)
        except TypeError:
            self.paramCtrl.SetValue(str(value))
        except AttributeError:
            self.paramCtrl.SetPath(value)


class ReflexBaseGui(wx.App):
    def __init__(self, io, dm):
        self.io = io
        self.dm = dm
        wx.App.__init__(self)

    def OnInit(self):
        respath = os.path.join(os.path.dirname(__file__), 'reflexgui.xrc')
        # changed global size, filesList size and removed notebook
        #respath = os.path.join(os.path.dirname(reflex.__file__),
        #                       'reflex_interactive_gui.xrc')
        self.res = xrc.XmlResource(respath)
        self.frame = self.res.LoadFrame(None, 'ReflexInteractiveWxApp')
        if not self.frame:
            raise Exception("Frame not found")

        if hasattr(self.dm, 'setWindowTitle'):
            self.frame.SetTitle(self.dm.setWindowTitle())

        self.plotPanel = xrc.XRCCTRL(self.frame, 'plotPanel')
        self.ctrlPanel = xrc.XRCCTRL(self.frame, 'ctrlPanel')
        self.ctrlPanel.SetScrollbars(0, 1, 0, 1)
        self.ctrlPanel.SetAutoLayout(1)
        self.statusBar = xrc.XRCCTRL(self.frame, 'statusBar')
        self.setDisableCheck = xrc.XRCCTRL(self.frame, 'setDisableCheck')

        self.Bind(wx.EVT_BUTTON, self.onCont, id=xrc.XRCID('contBtn'))
        self.Bind(wx.EVT_BUTTON, self.onRepeat, id=xrc.XRCID('repeatBtn'))
        self.Bind(wx.EVT_BUTTON, self.onHelp, id=xrc.XRCID('helpBtn'))
        self.Bind(wx.EVT_CHECKBOX, self.onSetDisable, self.setDisableCheck)
        self.Bind(wx.EVT_COLLAPSIBLEPANE_CHANGED, self.onFileInfoPaneChanged)
        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.onItemSelected)

        self.dm.readFitsData(self.io.inputs.in_sof.files +
                             self.io.inputs.in_sof_rec_orig.files)
        self.datasetInfoText = xrc.XRCCTRL(self.frame, 'datasetInfoText')
        self.datasetInfoText.SetLabel("This data belongs to dataset:\n%s" %
                                      self.io.inputs.in_sof.datasetName)
        self.filesInfo       = xrc.XRCCTRL(self.frame, 'filesInfo')
        self.filesList       = xrc.XRCCTRL(self.frame, 'filesList')
        self.setInitSopCheck = xrc.XRCCTRL(self.frame, 'setInitSopCheck')
        if(not self.io.is_init_sop_enable()):
            self.setInitSopCheck.Hide()

        self.setupParameters()
        self.createPlots()
        self.setFilesInfo()
        # normally triggerd by Collapse but that may not run on start on mac
        # which causes broken parameter layout
        self.ctrlPanel.FitInside()
        self.frame.Show()
        return True

    @classmethod
    def show(self, io, dataplotmanager):
        app = ReflexBaseGui(io, dataplotmanager)
        app.MainLoop()
        return app

    def createPlots(self):
        self.figure = Figure()
        self.canvas = FigCanvas(self.plotPanel, wx.ID_ANY, self.figure)
        self.canvas.mpl_connect('motion_notify_event', self.onMotion)
        self.canvas.mpl_connect('axes_enter_event', self.onEnterAxes)
        self.canvas.mpl_connect('axes_leave_event', self.onLeaveAxes)

        self.toolbar = NavigationToolbar(self.canvas)
        self.frame.SetToolBar(self.toolbar)
        self.plotPanel.Bind(wx.EVT_SIZE, self.onResize)
        self.plotPanel.Bind(wx.EVT_IDLE, self.onIdle)
        self.canvas.SetSize(self.plotPanel.GetSize())
        self.needs_resize = False

        self.dm.addSubplots(self.figure)
        self.dm.frame = self.frame
        self.dm.plotProductsGraphics(self.figure, self.canvas)

    def onResize(self, event):
        self.needs_resize = True

    def onMotion(self, event):
        if event.inaxes:
            self.statusBar.SetStatusText(event.inaxes.format_coord(event.xdata,
                                                                   event.ydata))
        else:
            self.statusBar.SetStatusText((''), 0)

    def onEnterAxes(self, event):
        if hasattr(event.inaxes, 'tooltip'):
            tip = self.canvas.GetToolTip()
            if not tip:
                tip = wx.ToolTip(tip=event.inaxes.tooltip)
                self.canvas.SetToolTip(tip)
            else:
                tip.SetTip(event.inaxes.tooltip)
            tip.Enable(True)

    def onLeaveAxes(self, event):
        try:
            self.canvas.GetToolTip().Enable(False)
        except:
            pass

    def onIdle(self, event):
        if self.needs_resize:
            self.canvas.SetSize(self.plotPanel.GetSize())
            self.needs_resize = False

    def onFileInfoPaneChanged(self, event):
        self.filesInfo.GetParent().Layout()
        self.ctrlPanel.FitInside()

    def onItemSelected(self, event):
        clipData = wx.TextDataObject()
        clipData.SetText(self.all_fitsFiles[self.filesList.GetFocusedItem()].name)
        wx.TheClipboard.Open()
        wx.TheClipboard.SetData(clipData)
        wx.TheClipboard.Close()

    def setFilesInfo(self):
        # filesInfo does not get proper class on wxPython3 on macos
        if hasattr(self.filesInfo, 'Collapse'):
            self.filesInfo.Collapse()
        self.filesList.InsertColumn(0, "Category")
        self.filesList.InsertColumn(1, "File Name")
        self.all_fitsFiles = self.io.inputs.in_sof.files + \
                             self.io.inputs.in_sof_rec_orig.files
        for file in self.all_fitsFiles:
            self.filesList.Append([file.category, file.name])
            self.filesList.SetColumnWidth(0, wx.LIST_AUTOSIZE)
            self.filesList.SetColumnWidth(1, wx.LIST_AUTOSIZE)
            self.filesList.SetToolTipString("List of files that have been an"
                "input of this window. If you click on a particular file, you"
                " can double click or press Ctrl-C to copy the full path name"
                " onto the clipboard.")

    def setupParameters(self):
        parTab = xrc.XRCCTRL(self.frame, 'parameterPanel')
        parTab.SetMinSize((170, 340))
        parGrid = wx.FlexGridSizer(cols=2, vgap=0, hgap=0)
        parGrid.AddGrowableCol(1)
        parTab.SetSizer(parGrid, 1)

        sop_params = dict()
        for p in self.io.inputs.in_sop:
            sop_params[(p.recipe, p.displayName)] = p.value

        self.paramWidgets = dict()
        for p in self.dm.setInteractiveParameters():
            if not (p.recipe, p.name) in sop_params:
                continue

            p.value = \
                VisirParameter.to_native(p.type,
                                         sop_params[(p.recipe, p.name)])
            w = ParamWidget(parTab, p)
            parGrid.Add(w.paramLabel,
                        flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL, border=5)
            parGrid.Add(w.paramCtrl,
                        flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL | wx.EXPAND,
                        border=5)
            self.paramWidgets[p.name] = w
        self.dm.parWidgets = self.paramWidgets

    def onCont(self, event):
        changed = []
        for p in self.paramWidgets.values():
            if not p.is_default():
                changed.append(p.param.name)
        if changed:
            if len(changed) > 1:
                s  = 'Parameters "%s" have been changed.' % ','.join(changed)
            else:
                s  = 'Parameter" %s" has been changed.' % changed[0]
            dlg = wx.MessageDialog(self.frame,
                    "%s To take effect the recipe must be rerun. "
                    "Discard changes and continue?" % s,
                                   "Confirm",
                                   wx.OK | wx.CANCEL)
            result = dlg.ShowModal()
            dlg.Destroy()
            if result != wx.ID_OK:
                return
        self.io.outputs.iteration_complete = 'true'
        self.io.outputs.out_sof = self.io.inputs.in_sof
        self.io.outputs.out_sop = self.io.inputs.in_sop

        if hasattr(self, 'set_next_iteration_enable'):
            self.io.outputs.set_enable = self.set_next_iteration_enable

        self.frame.Close(True)

    def verifyParameters(self):
        params = dict()
        for p in self.paramWidgets.values():
            params[p.param.name] = dict(value=p.get_value(),
                                        isdefault=p.is_default())
        s = None
        if hasattr(self.dm, 'verifyParameters'):
            s = self.dm.verifyParameters(params)
        return s

    def onRepeat(self, event):
        s = self.verifyParameters()
        if s:
            dlg = wx.MessageDialog(self.frame, s + '\nIgnore and continue?' ,
                                   "Confirm", wx.OK | wx.CANCEL)
            result = dlg.ShowModal()
            dlg.Destroy()
            if result != wx.ID_OK:
                return
        self.io.outputs.out_sop_loop = list()
        for  p in self.paramWidgets.values():
            param = RecipeParameter(displayName=p.param.name,
                                    value=p.get_value(), recipe=p.param.recipe)
            self.io.outputs.out_sop_loop.append(param)

        self.io.outputs.iteration_complete = 'false'
        self.io.outputs.out_sof_loop = self.io.inputs.in_sof_rec_orig
        if hasattr(self, 'set_next_iteration_enable'):
            self.io.outputs.set_enable = self.set_next_iteration_enable

        self.frame.Close(True)

    def onSetDisable(self, event):
        if self.setDisableCheck.GetValue():
            self.set_next_iteration_enable = 'false'
        else:
            self.set_next_iteration_enable = None

    def onHelp(self, event):
        if hasattr(self.dm, 'setWindowHelp'):
            window_help_msg = self.dm.setWindowHelp()
        else:
            window_help_msg = ''
        general_help_msg = """

    The window has several parts:\n
       1. Plot area. This area shows how good the reduction was performed
       2. Parameter area. This area shows the parameters used for the execution of the recipe. The parameters can be changed if the recipe has to be re-run with new parameters
       3. Button area. These buttons control the interactivity of the window\n
            a) Continue wkf. This button sends the current results to the next recipe in the workflow.\n
            b) Re-run recipe. This button will execute the predict recipe again with the new parameters\n
            c) Help. This button shows this help
       4. Top toolbar area. These buttons allow certain interactivity with the plots (zoom, shift, layout) as well as exporting to png
            """
        dlg = wx.MessageDialog(self.frame, window_help_msg + general_help_msg,
                               "Help",
                               wx.OK | wx.ICON_INFORMATION)
        dlg.ShowModal()
        dlg.Destroy()



from reflex import RecipeParameter, FitsFile

_ptypes = dict(file=str, list=str, string=str,
               float=float, int=int,
               bool=lambda x: True if x.lower() == 'true' else False)

class VisirParameter:

    def __init__(self, recipe, name, type,
                 description="", display_name=None,
                 options=[]):
        if type not in _ptypes:
            raise ValueError('%s is invalid type' % type)
        if type == 'list' and not options:
            raise ValueError('List parameter needs set of options')
        self.recipe = recipe
        self.display_name = name if not display_name else display_name
        self.name = name
        self.type = type
        self.description = description
        self.options = options

    @staticmethod
    def to_native(ptype, value):
        if not isinstance(value, basestring):
            return value
        if ptype in _ptypes:
            return _ptypes[ptype](value)
        return str(value)

    @staticmethod
    def to_string(ptype, value):
        if ptype == 'bool':
            return 'TRUE' if value else 'FALSE'
        return str(value)

def parseSop(sop):
    """
    remove when in reflex.py
    """
    retVal = list()
    sop_splitted = sop.split(',')
    for parameter in sop_splitted:
        param_recipe = parameter.partition('=')[0].partition(':')[0]
        param_name = parameter.partition('=')[0].partition(':')[2]
        param_value = parameter.partition('=')[2]
        p = RecipeParameter(param_recipe, param_name, "", "")
        p.value = param_value
        retVal.append(p)
    return retVal


try:
    from reflex import parseSof
except:
    # copy to avoid hard pyfits dependency
    def parseSof(sof):
        """
        parses the SoF passed by the PythonActor to 
        a list of FitsFile objects
        """
        retVal = list()
        dataset = sof[:sof.index('|')]
        only_sof = sof[sof.index('|')+1:]
        if len(only_sof) == 0 : 
          return retVal, dataset
        sof_splitted = only_sof.split(',')
        for frame in sof_splitted:
            name     = frame[0:frame.find(";")]
            category = frame[(frame.find(";")+1):(frame.rfind(";"))]
            purpose  = frame[(frame.rfind(";"))+1:]
            retVal.append(FitsFile(name, category, purpose))
        return retVal, dataset


def serialize_par(par):
    return "%s:%s=%s" % (par.recipe, par.name, str(par.value))


def join_sop(parlist):
    return ",".join((serialize_par(p) for p in parlist))


def serialize_fits(fits):
    return "%s;%s;%s" % (fits.name, fits.category, "")#fits.purpose)


def join_sof(fitslist, dataset):
    s = dataset + "|"
    s += ",".join((serialize_fits(f) for f in fitslist))
    return s




from __future__ import with_statement
import sys
from threading import Condition, Thread
from functools import wraps


def percentile(d, q):
    import numpy
    try:  # requires np >= 1.5.0, uses weighted percentile
        return numpy.percentile(d, q)
    except AttributeError:
        sd = numpy.sort(d.ravel())
        return [sd[int(v / 100. * d.size)] for v in q]


def synchronized(tlockname):
    """A decorator to place an instance based lock around a method """

    def _synched(func):
        @wraps(func)
        def _synchronizer(self, *args, **kwargs):
            tlock = getattr(self, tlockname)
            tlock.acquire()
            try:
                return func(self, *args, **kwargs)
            finally:
                tlock.release()
        return _synchronizer
    return _synched


class Future:
    def __init__(self, func, args=[], callback=None, data=None):
        self.__done = 0
        self.__result = None
        self.__status = 'working'
        self.__excpt = None
        self.__callback = callback
        self.data = data

        # Notify on this Condition when result is ready
        self.__C = Condition()

        # Run the actual function in a separate thread
        self.__T = Thread(target=self.Wrapper, args=[func] + list(args))
        self.__T.setName("FutureThread")
        self.__T.start()

    def __repr__(self):
        return '<Future at ' + hex(id(self)) + ':' + self.__status + '>'

    def get(self):
        self.__C.acquire()
        while self.__done == 0:
            self.__C.wait()
        self.__C.release()
        if self.__excpt:
            raise self.__excpt[0], self.__excpt[1], self.__excpt[2]
        return self.__result

    def Wrapper(self, func, *args):
        # Run the actual function, and let us housekeep around it
        self.__C.acquire()
        try:
            self.__result = func(*args)
        except:
            self.__result = "Exception raised within Future"
            self.__excpt = sys.exc_info()

        self.__done = 1
        self.__status = repr(self.__result)
        self.__C.notify()
        self.__C.release()
        if self.__callback:
            self.__callback(self)


def log_invocation(fn='/tmp/reflex_invoc.log'):
    """ log invocation of actor for easy reexecution with:
        eval $(cat /tmp/reflex_invoc.log)
    """
    import os
    try:
        with open(fn, 'w') as f:
            path = ""
            s = " ".join(['"%s"' % x for x in sys.argv])
            try:
                import reflex
                path = '"%s:"' % os.path.dirname(reflex.__file__)
            except ImportError:
                pass
            f.write('PYTHONPATH=%s$PYTHONPATH python -W all %s\n' % (path, s))
    except:
        pass



from matplotlib.patches import RegularPolygon, Rectangle
from matplotlib.widgets import RectangleSelector


class SquareSelector(RectangleSelector):
    # smaller minspan makes little sense due to interpolation radius
    def __init__(self, ax, onselect=None, useblit=True, rectprops=None,
                 dflt_size=50, minspan=15, maxselections=4):
        if rectprops is None:
            rectprops = dict(facecolor='red', edgecolor='black',
                             alpha=0.5, fill=True)
        self.sq_onselect = onselect
        RectangleSelector.__init__(self, ax=ax, onselect=lambda x, y: True,
                                   useblit=useblit, rectprops=rectprops,
                                   minspanx=None, minspany=None,
                                   spancoords='pixels', drawtype='box')
        self.selections = []
        self.dflt_size = dflt_size
        self.minspan = minspan
        self.maxselections = maxselections
        self.cross = None

    def update(self):
        if self.useblit:
            if self.background is not None:
                self.canvas.restore_region(self.background)
            self.ax.draw_artist(self.to_draw)
            if self.cross is not None:
                self.ax.draw_artist(self.cross)
            for a in self.selections:
                self.ax.draw_artist(a[0])
                self.ax.draw_artist(a[1])
            self.canvas.blit(self.ax.bbox)
        else:
            self.canvas.draw_idle()
        return False

    def press(self, event):
        # Is the correct button pressed within the correct axes?
        if self.ignore(event):
            return
        if event.button == 3:
            rem = []
            for s in reversed(self.selections):
                if s[1].contains(event)[0]:
                    x, y = s[1].xy
                    h = s[1].get_height() / 2
                    if self.sq_onselect is not None:
                        self.sq_onselect([x + h, y + h, h], False)
                    self.ax.patches.remove(s[0])
                    self.ax.patches.remove(s[1])
                    rem.append(s)
                    break
            for r in rem:
                self.selections.remove(r)
            self.canvas.draw_idle()
            return
        if len(self.selections) >= self.maxselections:
            return

        # make the drawed box/line visible get the click-coordinates,
        # button, ...
        self.to_draw.set_visible(self.visible)
        self.eventpress = event

        self.cross = RegularPolygon((event.xdata, event.ydata), numVertices=4,
                                    radius=1, facecolor='white')
        s = self.dflt_size
        self.to_draw.set_width(s * 2)
        self.to_draw.set_height(s * 2)
        self.to_draw.set_x(event.xdata - s)
        self.to_draw.set_y(event.ydata - s)
        self.ax.add_patch(self.cross)
        self.eventpress.data = (event.xdata, event.ydata, s)
        self.update()
        return False

    def release(self, event):
        if not (self.eventpress is None or self.ignore(event)) and \
           event.button == 1 and len(self.selections) < self.maxselections:
            c, r = draw_selection(*self.eventpress.data)
            self.ax.add_patch(r)
            self.ax.add_patch(c)
            self.selections.append((c, r))
            self.sq_onselect(self.eventpress.data, True)

        if self.cross is not None:
            self.ax.patches.remove(self.cross)
            self.cross = None

        if self.ignore(event):
            self.to_draw.set_visible(False)
            self.canvas.draw_idle()

        RectangleSelector.release(self, event)

    def onmove(self, event):
        if self.eventpress is None or self.ignore(event):
            return
        x, y = event.xdata, event.ydata

        bx, by = self.eventpress.xdata, self.eventpress.ydata
        s = max(abs(by - y), abs(bx - x))
        if s < self.minspan:
            s = self.dflt_size

        self.to_draw.set_width(s * 2)
        self.to_draw.set_height(s * 2)

        self.to_draw.set_x(bx - s)
        self.to_draw.set_y(by - s)

        self.eventpress.data = (bx, by, s)
        self.update()
        return False

    def get_selections(self):
        res = []
        for s in self.selections:
            x, y = s[1].xy
            w = s[1].get_height()
            res.append((x + w / 2, y + w / 2, w))
        return res


def draw_selection(x, y, s, rectprops=None):
    if rectprops is None:
        rectprops = dict(facecolor='white', edgecolor='black',
                         alpha=0.5, fill=True)

    cross = RegularPolygon((x, y), numVertices=4, radius=1, facecolor='white')
    rect = Rectangle((x - s, y - s), s * 2, s * 2, visible=True, **rectprops)
    return cross, rect



from __future__ import with_statement
from visir_utils import synchronized, percentile
from threading import Lock
import numpy
import matplotlib.colors


def freedman_bins(d, max_bins=1000):
    """ bin width via Freedman-Diaconis rule """
    # np.percentile only available in np >= 1.5
    n = d.size
    dsorted = numpy.sort(d.ravel())
    v25 = dsorted[n / 4 - 1]
    v75 = dsorted[(3 * n) / 4 - 1]
    dx = 2 * (v75 - v25) / (n ** (1. / 3.))
    nbins = numpy.ceil((dsorted[-1] - dsorted[0]) * 1. / dx)
    return min(max_bins, max(1, nbins))


def limit_cscale(d, q):
    """ imshow(d, **limit_cscale(d, (2, 98)))"""
    vmin, vmax = percentile(d, q)
    return {'vmax': vmax, 'vmin': vmin}


class ScaleNormalize(matplotlib.colors.Normalize):
    def __init__(self, scale=None, vmin=None, vmax=None, clip=False):
        matplotlib.colors.Normalize.__init__(self, vmin, vmax, clip)
        self.scale = scale

    def autoscale_None(self, A):
        matplotlib.colors.Normalize.autoscale_None(self, A)
        if self.scale and not self.scaled():
            self.vmin = self.scale(self.vmin)
            self.vmax = self.scale(self.vmax)

    def __call__(self, val, clip=None):
        if self.scale:
            val = self.scale(val)
        return matplotlib.colors.Normalize.__call__(self, val, clip)


class BasePlot:
    def __init__(self, figure, ax, data=[], tooltip=''):
        self.lock = Lock()
        self.figure = figure
        self.ax = ax
        self.ax.tooltip = tooltip
        self.data = data
        self.norm = None

    def inaxes(self, *args):
        for e in args:
            if e.inaxes != self.ax:
                return False
            x, y = e.xdata, e.ydata
            xl, xh = self.ax.get_xbound()
            yl, yh = self.ax.get_ybound()
            if x < xl or x >= xh or y < yl or y >= yh:
                return False
        return True

    @synchronized('lock')
    def set_data(self, data):
        self.data = data

    def set_loading(self, txt='Loading'):
        self.ax.cla()
        self.ax.set_title(txt)
        self.figure.canvas.draw()


class LinePlot(BasePlot):
    def __init__(self, figure, ax, ydata=[], xdata=[], tooltip='',
                 *args, **kwargs):
        if not xdata:
            xdata = range(len(ydata))
        data = numpy.vstack([ydata, xdata])
        BasePlot.__init__(self, figure, ax, data, tooltip)
        self.args = args
        self.kwargs = kwargs

    @staticmethod
    def conv_data(ydata, xdata=[]):
        if len(xdata) == 0:
            xdata = numpy.arange(len(ydata))
        return numpy.vstack([ydata, xdata])

    def draw(self, *args, **kwargs):
        with self.lock:
            if isinstance(self.data, list) or self.data.shape[0] != 2:
                self.data = self.conv_data(self.data)
        ydata = self.data[0]

        self.args = args
        self.kwargs = kwargs
        if self.ax.lines:
            self.ax.lines[0].set_data(self.data[1], ydata)
            self.ax.relim()
            self.ax.autoscale_view(False, True, True)
        else:
            self.ax.plot(self.data[1], ydata, *self.args, **self.kwargs)
        self.figure.canvas.draw()


class ImagePlot(BasePlot):
    def __init__(self, figure, ax, data=[], tooltip='', *args, **kwargs):
        BasePlot.__init__(self, figure, ax, data, tooltip)
        self.args = args
        self.kwargs = kwargs

    @synchronized('lock')
    def set_norm(self, norm):
        if norm and self.ax.images:
            ll, hl = self.ax.images[0].get_clim()
            norm.vmin = norm.scale(ll)
            norm.vmax = norm.scale(hl)
        self.norm = norm

    @synchronized('lock')
    def set_clim(self, v):
        if self.ax.images:
            if self.norm:
                self.ax.images[0].set_clim([self.norm.scale(x) for x in v])
            else:
                self.ax.images[0].set_clim(v)

    def draw(self, *args, **kwargs):
        rescale = kwargs.pop('rescale', False)
        self.args = args
        self.kwargs = kwargs
        # work around matplotlib < 1.2.0 issue 1005
        if self.data.dtype.isnative:
            data = self.data
        else:
            data = self.data.byteswap().newbyteorder()
        if self.ax.images:
            img = self.ax.images[0]
            img.set_data(data)
            self.ax.relim()
            if self.norm:
                img.set_norm(self.norm)
            else:
                ll, hl = img.get_clim()
                img.set_norm(matplotlib.colors.Normalize(vmin=ll, vmax=hl))
            if rescale:
                self.ax.autoscale_view(True, True, True)
        else:
            self.ax.imshow(data, *self.args, **self.kwargs)
        self.figure.canvas.draw()



from SimpleXMLRPCServer import SimpleXMLRPCServer
from SimpleXMLRPCServer import SimpleXMLRPCRequestHandler
import xmlrpclib
import pickle
try:
    import pyfits
except ImportError:
    import astropy.io.fits as pyfits
# improves performance a few percent
pyfits.ENABLE_RECORD_VALUED_KEYWORD_CARDS = False

import numpy
import cStringIO
def ser_array(x):
    if x is None:
        return x
    f = cStringIO.StringIO()
    numpy.lib.format.write_array(f, x)
    return xmlrpclib.Binary(f.getvalue())

def deser_array(s):
    if s == None:
        return None
    f = cStringIO.StringIO(s.data)
    return numpy.lib.format.read_array(f)


# Restrict to a particular path.
class RequestHandler(SimpleXMLRPCRequestHandler):
        rpc_paths = ('/RPC2',)

class Server:
    def __init__(self, port=8000):
        # Create server
        self.server = SimpleXMLRPCServer(("localhost", port),
                                         requestHandler=RequestHandler,
                                         allow_none=True,
                                         logRequests=False)
        self.server.register_introspection_functions()
        self.server.register_multicall_functions()
        self.server.register_function(self.quit)

        self.server.register_function(self.pyfits_open, 'pyfits_open')
        self.server.register_function(self.pyfits_close, 'pyfits_close')

        self.server.register_function(self.pyfits_getheader, 'pyfits_getheader')
        self.server.register_function(self.pyfits_getdata, 'pyfits_getdata')
        self.server.register_function(self.pyfits_getval, 'pyfits_getval')
        self.server.register_function(self.pyfits_custom, 'pyfits_custom')
        self.server.register_function(self.len, 'len')

        self.files = dict()

        self._run = True

    def run(self):
        while self._run:
            self.server.handle_request()

    def quit(self):
        self._run = False
        return 0

    def pyfits_open(self, fn):
        # while memory mapping would be good it is handled badly by pyfits
        # see tickets 175 and 177
        if fn not in self.files:
            # workaround pyfits <= 2.1.1 unicode issue (CNSHD750437)
            try:
                self.files[fn] = pyfits.open(fn, memmap=False)
            except AttributeError:
                self.files[fn] = pyfits.open(str(fn), memmap=False)

    def pyfits_close(self, fn):
        if fn in self.files:
            del self.files[fn]

    def pyfits_getheader(self, fn, idx):
        self.pyfits_open(fn)
        f = self.files[fn]
        if isinstance(idx, int):
            return dict(f[idx].header)
        if isinstance(idx, list):
            vals = []
            for i in idx:
                vals.append(dict(f[i].header))
            return vals
        if idx is None:
            return [dict(x.header) for x in f]

    def pyfits_getdata(self, fn, idx):
        self.pyfits_open(fn)
        f = self.files[fn]
        if isinstance(idx, int):
            return ser_array(f[idx].data)
        if isinstance(idx, list):
            vals = []
            for i in idx:
                vals.append(ser_array(f[i].data))
            return vals
        if idx is None:
            return [ser_array(x.data) for x in f]
        return ser_array(f[idx].data)

    def pyfits_getval(self, fn, keyword, idx, skip_first=True):
        self.pyfits_open(fn)
        f = self.files[fn]
        if isinstance(idx, int):
            return f[idx].header.get(keyword)
        if isinstance(idx, list):
            vals = []
            for i in idx:
                vals.append(f[i].header.get(keyword))
            return ser_array(numpy.array(vals))
        if idx is None:
            if skip_first:
                return ser_array(numpy.array([x.header.get(keyword) for x in f[1:]]))
            else:
                return ser_array(numpy.array([x.header.get(keyword) for x in f]))

    def pyfits_custom(self, fn, pfun, args=None):
        self.pyfits_open(fn)
        f = self.files[fn]
        fun = pickle.loads(pfun)
        if args:
            res = fun(f, *args)
        else:
            res = fun(f)
        if isinstance(res, numpy.ndarray):
            return ser_array(res)
        else:
            return res

    def len(self, fn):
        return self.pyfits_custom(fn, len)

if __name__ == "__main__":
    s = Server()
    s.run()



from __future__ import with_statement
# must succeed, min requirement py 2.5
import sys
from collections import defaultdict
from visir_utils import log_invocation, Future
import types
log_invocation()

# import the needed modules
try:
    import numpy
    try:
        import pyfits
    except ImportError:
        import astropy.io.fits as pyfits
    # needs pyfits
    import fitsloader
    import matplotlib
    from matplotlib.widgets import RadioButtons, Button
    matplotlib.use('WXAgg')
    # import before wx to ensure we get 2.8 even if 2.6 is also installed
    import matplotlib.backends.backend_wxagg
    from matplotlib.widgets import SpanSelector

    from plot_utils import LinePlot, ImagePlot, ScaleNormalize, freedman_bins
    from pipeline_display import ScatterDisplay
    from reflexlib import VisirParameter
    from reflexgui import ReflexBaseGui, PlotDialog, BeamDialog

    import wx

    import_success = True
except (ImportError, NameError), e:
    import_success = False
    print "Error importing modules: ", str(e)


def myspanignore(self, event):
    # ignore in pan and zoom mode
    # https://github.com/matplotlib/matplotlib/issues/1368
    if not self.canvas.widgetlock.available(self):
        return True
    return SpanSelector.ignore(self, event)


def get_keys_wb(keyword, fl, files):
    cor = fl.getval_all(keyword, files=files, skip_first=True)
    bounds = numpy.cumsum([x.size for x in cor])
    cor = numpy.concatenate(cor)
    return cor, [(0, 0)] + [(x, y) for x, y in zip(bounds, files)]


def get_slice(fits, y, x):
    vals = []
    for ext in fits[1:]:
        vals.append(float(ext.data[int(y), int(x)]))
    return vals


def get_beam_coords(fl, files):
    res = []
    for fn in files:
        bid = fl.getval(fn, 'ESO QC BEAMID', 0)
        x = fl.getval(fn, 'ESO QC BEAMX', 0)
        y = fl.getval(fn, 'ESO QC BEAMY', 0)
        res.append((bid, x, y))
    return res


class DataPlotterManager:
    stats = dict(
                 stdev=dict(key='ESO QC STDEV',
                            name='stdev', title='Standard deviation',
                            label='Std. deviation', unit='ADU / s'),
                 mad=dict(key='ESO QC MAD', name='mad',
                          title='Median absolute deviation',
                          label='MAD', unit='ADU / s'),
                 median=dict(key='ESO QC MEDIAN', name='median',
                             title='Median', label='Median', unit='ADU / s'),
                 fwhmx=dict(key='ESO QC FWHMX', name='fwhmx', unit='pixel',
                            title='Full width half maximum X', label='FWHM X'),
                 fwhmy=dict(key='ESO QC FWHMY', name='fwhmy', unit='pixel',
                            title='Full width half maximum Y', label='FWHM Y'),
                 correl=dict(key='ESO QC CORR', name='correl',
                             title='Cross-correlation', label='Correlation'),
                )
    dflt_stat = 'correl'
    SHOW_STATS = 1
    SHOW_SLICE = 2

    def __init__(self, fitsloader):
        self.fl = fitsloader
        self.dlg = None
        self.beams = None
        self.undef_beams = False

    def readFitsData(self, fitsFiles):
        #Read all the products
        fits = defaultdict(list)
        for frame in fitsFiles:
            if frame == '':
                continue
            category = frame.category
            fits[category].append(frame.name)
        self.fits = fits

        self.templates = []
        for f in fits['CORRELATION_TEMPLATE']:
            # workaround pyfits <= 2.1.1 unicode issue (CNSHD750437)
            try:
                self.templates.extend(pyfits.open(f))
            except AttributeError:
                self.templates.extend(pyfits.open(str(f)))

        try:
            self.mean = pyfits.getdata(fits['IMG_REPACKED_MEAN'][0], 0)
        except:
            self.mean = None

    def addSubplots(self, figure):
        if len(self.fits['BEAM_DETECTED']) > 0:
            self.figure = figure
            self.plt = dict()
            self.plt['av'] = ImagePlot(figure, figure.add_subplot(334))
            self.plt['img'] = ImagePlot(figure, figure.add_subplot(335))
            self.plt['tmpl'] = ImagePlot(figure, figure.add_subplot(336))
            self.plt['shifts'] = ImagePlot(figure, figure.add_subplot(313))
            self.plt['stats'] = LinePlot(figure, figure.add_axes([0.30, 0.70,
                                                                 0.675, 0.25]))
            self.rax = figure.add_axes([0.02, 0.75, 0.20, 0.13])
            self.rax.tooltip = 'Select statistics to plot'
            self.statskeys = sorted(self.stats.keys())
            self.statslabels = [self.stats[x]['label'] for x in self.statskeys]
            self.rad = RadioButtons(self.rax, self.statslabels,
                                    active=self.statskeys.index(self.dflt_stat))
            self.bax = figure.add_axes([0.08, 0.65, 0.13, 0.05])
            self.beamselect = Button(self.bax, 'Select beams', color='gray')
            self.beamselect.on_clicked(self.select_beams)
        else:
            self.nodata = figure.add_subplot(111)

    def get_image(self, idx):
        bounds = [(0, 0), (1000, self.fits['BEAM_DETECTED'][0])]
        if idx != 0:
            bounds = self.statsdata[self.dflt_stat].get()[1]
        idx = int(idx) + 1
        for i, v in enumerate(bounds):
            if idx <= v[0]:
                return self.fl.getdata(v[1], idx - bounds[i - 1][0])
        return numpy.zeros((200, 200))

    def get_slice(self, files, x, y):
        res = []
        for fn in files:
            res.append(self.fl.custom(fn, get_slice, (x, y)))
        return  numpy.concatenate(res)

    def rad_clicked(self, label):
        self.plt['stats'].set_loading()
        # add_done_callback would be useful here
        label = self.statskeys[self.statslabels.index(label)]
        Future(lambda: self.statsdata[label].get(),
               callback=lambda f: wx.CallAfter(self.plot_stats, f.get(), label))
        self.showing = self.SHOW_STATS

    def change_clim(self, vmin, vmax, plt=None):
        if plt is None or not plt.ax.images:
            return
        plt.set_clim([vmin, vmax])
        plt.figure.canvas.draw_idle()

    def detach(self, plt):
        if self.figure.canvas.HasCapture():
            self.figure.canvas.ReleaseMouse()
        dlg = PlotDialog(self.frame, title=plt.ax.get_title(),
                         size=wx.Size(700, 700))
        dlg.plt = ImagePlot(dlg.figure, dlg.figure.add_subplot(111))
        dlg.plt.set_data(plt.data)
        dlg.plt.draw(interpolation='nearest', origin='lower')
        dlg.plt.figure.colorbar(dlg.plt.ax.images[0])
        dlg.plt.detached = True
        dlg.on_click = lambda e: wx.CallAfter(self.show_popup, dlg.plt) \
                                 if e.button == 3 else None
        dlg.figure.canvas.mpl_connect('button_press_event', dlg.on_click)

        # close scale dialog
        def dlgclose(event):
            dlg.Destroy()
            if self.dlg:
                self.dlg.Destroy()
                self.dlg = None
        dlg.Bind(wx.EVT_CLOSE, dlgclose)
        dlg.Bind(wx.EVT_BUTTON, dlgclose)

        dlg.Show()

    def scale_dialog(self, plt):
        if self.figure.canvas.HasCapture():
            self.figure.canvas.ReleaseMouse()
        if self.dlg is None:
            self.dlg = PlotDialog(self.frame, title='Colorscale')
        else:
            self.dlg.figure.clf()
        ax = self.dlg.figure.add_subplot(111)

        ax.hist(plt.data.ravel(), histtype='stepfilled',
                bins=freedman_bins(plt.data))
        self.dlg.span = SpanSelector(ax,
                                     lambda x, y : self.change_clim(x, y, plt),
                                     'horizontal', useblit=True,
                                     rectprops=dict(alpha=0.5, facecolor='red'))
        self.dlg.span.ignore = types.MethodType(myspanignore, self.dlg.span)

        self.dlg.figure.canvas.draw_idle()
        def dlgclose(event):
            self.dlg.Destroy()
            self.dlg = None
        self.dlg.Bind(wx.EVT_CLOSE, dlgclose)
        self.dlg.Bind(wx.EVT_BUTTON, dlgclose)
        self.dlg.Show()

    def select_beams(self, event):
        if self.figure.canvas.HasCapture():
            self.figure.canvas.ReleaseMouse()
        try:
            mheader = pyfits.getheader(self.fits['IMG_REPACKED_MEAN'][0], 0)
            pthrow = mheader['ESO TEL CHOP THROW'] / \
                     float(mheader['ESO INS PFOV'])
        except:
            pthrow = 100
        dlg = BeamDialog(self.frame, data=self.mean, dflt_size=pthrow / 2)
        ret = dlg.ShowModal()
        bp = ''
        for s in dlg.get_selections():
            x, y, w = s[1]
            sign = s[0]
            bp += '%s:%d,%d,%d;' % (sign, x, y, w)

        print bp, ret, wx.ID_OK
        if bp and ret == wx.ID_OK:
            self.parWidgets['beampos'].set_value(bp)
        dlg.Destroy()

    def show_popup(self, plt):
        try:
            if not hasattr(plt, 'popupmenu'):
                plt.popupmenu = wx.Menu()
                from functools import partial
                f = partial(self.OnPopupItemSelected, plt=plt)
                cmenu = wx.Menu()
                plt.popupmenu.AppendSubMenu(cmenu,'Color')
                pmenu = wx.Menu()
                plt.popupmenu.AppendSubMenu(pmenu, 'Colorscale')
                item = plt.popupmenu.Append(-1, 'Scale parameters')
                plt.figure.canvas.Bind(wx.EVT_MENU, f, item)

                if not hasattr(plt, 'detached'):
                    item = plt.popupmenu.Append(-1, 'Detach')
                    plt.figure.canvas.Bind(wx.EVT_MENU, f, item)

                for text in ('jet', 'gray', 'hot'):
                    item = cmenu.AppendRadioItem(-1, text)
                    plt.figure.canvas.Bind(wx.EVT_MENU, f, item)
                for text in ('linear', 'arcsinh', 'square', 'sqrt', 'log'):
                    item = pmenu.AppendRadioItem(-1, text)
                    plt.figure.canvas.Bind(wx.EVT_MENU, f, item)

            plt.figure.canvas.PopupMenu(plt.popupmenu)
            #self.popupmenu.Destroy()

            plt.draw()
        finally:
            # something does not release capture with popups, so force it
            if plt.figure.canvas.HasCapture():
                plt.figure.canvas.ReleaseMouse()

    def OnPopupItemSelected(self, event, plt=None):
        item = plt.popupmenu.FindItemById(event.GetId())
        v = item.GetText()
        if v == 'arcsinh':
            plt.set_norm(ScaleNormalize(numpy.arcsinh))
        elif v == 'square':
            plt.set_norm(ScaleNormalize(numpy.square))
        elif v == 'sqrt':
            plt.set_norm(ScaleNormalize(numpy.sqrt))
        elif v == 'log':
            plt.set_norm(ScaleNormalize(numpy.log))
        elif v == 'linear':
            plt.set_norm(None)
        elif v == 'Scale parameters':
            self.scale_dialog(plt)
        elif v == 'Detach':
            self.detach(plt)
        else:
            for im in plt.ax.images:
                im.set_cmap(getattr(matplotlib.cm, v, matplotlib.cm.jet))

    def onclick(self, event):
        if not self.figure.canvas.widgetlock.available(self):
            return
        if event.button == 3:
            plt = None
            for v in self.plt.values():
                if v.ax == event.inaxes and isinstance(v, ImagePlot) and \
                   v != self.plt['shifts']:
                    plt = v
                    break
            if plt:
                wx.CallAfter(self.show_popup, plt)

        elif self.showing == self.SHOW_STATS and \
            self.plt['stats'].inaxes(event) and \
            hasattr(self, 'cor') and \
            event.xdata >= 1 and round(event.xdata) < len(self.cor) + 1:
            x = int(round(event.xdata)) - 1
            img = self.get_image(x)
            try:
                ax = self.dlg.figure.axes[0]
                ax.cla()
                ax.hist(img.ravel(), histtype='stepfilled',
                        bins=freedman_bins(img))
                self.dlg.figure.canvas.draw()
            except:
                pass
            self.plot_image(img, x)

        elif self.plt['av'].inaxes(event):
            self.showing = self.SHOW_SLICE
            for c in self.rad.circles:
                c.set_facecolor(self.plt['stats'].ax.get_axis_bgcolor())
            x, y = int(event.ydata), int(event.xdata)
            self.plt['stats'].set_loading()
            Future(self.get_slice, args=(self.fits['BKG_CORRECTED'], x, y),
                   callback=lambda f: wx.CallAfter(self.plot_slice,
                                                   f.get(), x, y))

    def report_pixel(self, plt, x, y):
        xorig, yorig = x, y
        if hasattr(plt, 'scale_coord'):
            x, y = plt.scale_coord(x, y)
        if numpy.array(plt.data).size and \
           y > 0 and x > 0 and \
           x < plt.data.shape[1] and y < plt.data.shape[0]:
            return "x=%s y=%s value=%s" % (plt.ax.format_xdata(xorig),
                                           plt.ax.format_ydata(yorig),
                                           plt.ax.format_ydata(plt.data[int(y), int(x)]))
        else:
            return "x=%s y=%s" % (plt.ax.format_xdata(xorig),
                                  plt.ax.format_ydata(yorig))

    def plotProductsGraphics(self, figure, canvas):
        if len(self.fits['BEAM_DETECTED']) == 0:
            #Data not found info
            self.nodata.set_axis_off()
            self.text_nodata = \
            """Data not found. Possibly all images have been rejected"""
            self.nodata.text(0.1, 0.6, self.text_nodata,
                                     color='#11557c', fontsize=18, ha='left',
                                     va='center', alpha=1.0)
            self.nodata.tooltip = 'beam detection not found in the products'
            return

        from functools import partial
        get_bkeys = partial(get_keys_wb, fl=self.fl,
                            files=self.fits['BEAM_DETECTED'])

        plt_stats = self.plt['stats']
        plt_av = self.plt['av']
        plt_tmpl = self.plt['tmpl']
        plt_img = self.plt['img']

        plt_av.tooltip = 'Input image'
        try:
            self.plot_average(self.mean)
        except:
            plt_av.set_loading()
            Future(self.get_average,
                   callback=lambda f: wx.CallAfter(self.plot_average, f.get()))

        self.statsdata = dict()
        self.statsdata[self.dflt_stat] = \
            Future(get_bkeys, args=(self.stats[self.dflt_stat]['key'],),
                   callback=lambda f: wx.CallAfter(self.plot_stats,
                                                   f.get(), self.dflt_stat))
        for k in self.stats.keys():
            if k in self.statsdata:
                continue
            self.statsdata[k] = Future(get_bkeys, args=(self.stats[k]['key'],))

        plt_stats.set_loading()
        self.showing = self.SHOW_STATS

        plt_tmpl.set_data(self.templates[0].data)
        plt_tmpl.draw(interpolation='nearest', origin='lower')

        plt_img.tooltip = 'Plane'
        plt_img.set_loading()
        Future(lambda: self.get_image(0),
               callback=lambda f: wx.CallAfter(self.plot_image, f.get(), 0))

        plt_img.ax.format_coord = types.MethodType(self.report_pixel, plt_img)
        plt_av.ax.format_coord = types.MethodType(self.report_pixel, plt_av)
        plt_tmpl.ax.format_coord = types.MethodType(self.report_pixel, plt_tmpl)

        plt_tmpl.ax.set_title('Correlation template')
        plt_tmpl.ax.tooltip = 'Correlation template'

        self.plt['shifts'].set_loading()
        Future(lambda: (get_bkeys('CRPIX1')[0], get_bkeys('CRPIX2')[0]),
               callback=lambda f: wx.CallAfter(self.plot_shifts, f.get()))

        self.rad.on_clicked(self.rad_clicked)
        figure.canvas.mpl_connect('button_press_event', self.onclick)

    def add_beams(self, data):
        self.beams = data
        from matplotlib.patches import RegularPolygon
        for c in data:
            if c[0].upper() == 'UNDEFINED':
                self.undef_beams = True
                self.beamselect.color = 'red'
                self.bax.figure.canvas.draw_idle()
            p = RegularPolygon(c[1:], numVertices=4, alpha=0.2,
                               radius=self.plt['av'].data.shape[0] / 40,
                               facecolor='white')

            self.plt['av'].ax.add_patch(p)
        self.plt['av'].figure.canvas.draw_idle()

    def get_average(self):
        nimages = self.fl.custom(self.fits['BKG_CORRECTED'][0], len) - 1
        imginput = self.fl.getdata(self.fits['BKG_CORRECTED'][0],
                                   range(1, min(nimages, 100)))
        return numpy.average(numpy.array(imginput), axis=0)

    def plot_image(self, data, idx):
        plt_img = self.plt['img']
        plt_img.set_data(data)
        plt_img.ax.set_title('Image plane #%d' % (idx + 1))
        plt_img.draw(interpolation='nearest', origin='lower')

    def plot_average(self, data):
        plt_av = self.plt['av']
        plt_av.set_data(data)
        plt_av.ax.set_title('Input image')
        plt_av.draw(origin='lower', interpolation='nearest', rescale=False)

        Future(get_beam_coords, args=(self.fl, self.fits['BEAM_DETECTED']),
               callback=lambda f: wx.CallAfter(self.add_beams, f.get()))

    def plot_stats(self, data_, label):
        data, bounds = data_
        plt_stats = self.plt['stats']
        plt_stats.ax.set_title(self.stats[label]['title'] + ' (Click to view image)')
        plt_stats.ax.tooltip = 'Statistics of all planes'
        plt_stats.ax.set_xlabel('Plane number')
        if 'unit' in self.stats[label]:
            plt_stats.ax.set_ylabel('[%s]' % self.stats[label]['unit'])
        plt_stats.set_data(numpy.array((data, (numpy.arange(1, len(data) + 1)))))
        # ugly, better: disable click handles while loading
        self.cor = data
        label = self.statslabels[self.statskeys.index(label)]
        for i, l in enumerate(self.rad.labels):
            if l.get_text() == label:
                self.rad.circles[i].set_facecolor(self.rad.activecolor)
            else:
                self.rad.circles[i].set_facecolor(plt_stats.ax.get_axis_bgcolor())

        plt_stats.draw()

    def plot_slice(self, data, x, y):
        self.plt['stats'].set_data(numpy.array((data, (numpy.arange(1, len(data) + 1)))))
        self.plt['stats'].ax.set_title('Values over x=%d y=%d' % (x, y))
        self.plt['stats'].draw()

    def plot_shifts(self, scat):
        title_scatter = 'Shifts'
        tooltip_scatter = 'Shifts of the beams'
        scatx, scaty = scat
        # avoid crash on empty input
        if scatx.size == 0 or scaty.size == 0:
            scatx, scaty = [0], [0]
        scatterdisp = ScatterDisplay()
        scatterdisp.setLabels('X [pixel]', 'Y [pixel]')
        scatterdisp.display(self.plt['shifts'].ax, title_scatter,
                            tooltip_scatter, scatx, scaty)
        # add a histogram for many points
        if (len(scatx) > 50):
            self.plt['shifts'].data, binsx, binsy, aximg = \
                self.plt['shifts'].ax.hist2d(scatx, scaty, bins=25, alpha=0.75)
            aximg.set_cmap('gray')
            def shift_scale(x, y):
                # x, y transposed in data array
                return ((y - binsy[0]) / numpy.diff(binsy[:2]),
                        (x - binsx[0]) / numpy.diff(binsx[:2]))
            self.plt['shifts'].scale_coord = shift_scale
            self.plt['shifts'].ax.format_coord = types.MethodType(self.report_pixel, self.plt['shifts'])
        self.plt['shifts'].figure.canvas.draw()

    def setInteractiveParameters(self):
        rec = 'visir_util_detect_shift'
        paramList = [
            VisirParameter(rec, name='template', type='file',
                           description='Correlation template used to detect shifts and reject bad images. If none given an averaged image is used.'),
            VisirParameter(rec, name='naverage', type='int',
                           description='Number of planes to average before attempting to detect the shifts.'),
            VisirParameter(rec, name='no-reject', type='bool',
                           description='If true images with bad statistics will not be rejected and no shift correction is done.'),
            VisirParameter(rec, name='no-shift', type='bool',
                           description='Whether to apply the determined shifts in following recipes. Set to TRUE of shifts appear too large'),
            VisirParameter(rec, name='min-correlation', type='float',
                           description='Minimal correlation to the template required to be accepted.'),
            VisirParameter(rec, name='max-shift', type='float',
                           description='Maximal allowed object shift from the template for an image to be accepted.'),
            VisirParameter(rec, name='max-mad', type='float',
                           description='Maximal allowed median absolute deviation for an image to be accepted. <= 0 equals no limit.'),
            VisirParameter(rec, name='beampos', type='string',
                           description='Set via the "Select beams" button.'),
            VisirParameter(rec, name='method', type='list',
                           options=['xcorrelate', 'brightest'], description="""\
Method to determine beam shifts:
xcorrelate: use the cross correlation shift.
brightest: use the position of the brightest pixel"""),
        ]
        return paramList

    def verifyParameters(self, params):
        if params['no-reject']['value'].upper() == 'TRUE':
            if any((not params['max-mad']['isdefault'],
                    not params['min-correlation']['isdefault'],
                    not params['max-shift']['isdefault'])):
                    return 'Rejection parameters changed but "no-reject" is ' \
                      'still set. In order to reject frames this must be unset.'
        return None


#This is the 'main' function
if __name__ == '__main__':
    # for debugging

    # Create commandline interface
    from reflex_interactive_app import PipelineInteractiveApp
    app = PipelineInteractiveApp()
    app.parse_args()

    #Open the interactive window if enabled
    try:
        if app.isGUIEnabled() and import_success:
            # hard dependency on wx, has own io parsing
            #Get the specific functions for this window
            with fitsloader.get_fitsloader() as fl:
                dataPlotManager = DataPlotterManager(fl)
                dataPlotManager.in_pars = app.inputs.in_sop
                ReflexBaseGui.show(app, dataPlotManager)
        else:
            app.passProductsThrough()
    except (ImportError, NameError) as e:
        print 'Error:', e
        app.passProductsThrough()

    app.print_outputs()
    # prevents hang when using the threading module in fitsloader
    sys.exit(0)



# -*- coding: UTF-8 -*-
from __future__ import with_statement
from fitsrpcserver import Server, deser_array
from visir_utils import Future
import xmlrpclib
import pickle
from threading import Lock
try:
    from contextlib import contextmanager
except ImportError:
    pass

try:
    import multiprocessing as mp
except ImportError:
    # for python < 2.6
    print 'import of multiprocessing failed, falling back to threading'
    import threading
    import Queue
    threading.Process = threading.Thread
    # the GIL makes use of more threads pointless
    threading.cpu_count = lambda : 1
    threading.Queue = lambda : Queue.Queue()
    mp = threading

def run_server(port, q):
    try:
        s = Server(port=port)
        q.put(True)
        s.run()
    except:
        q.put(False)

try:
    @contextmanager
    def get_fitsloader():
        f = FitsLoader()
        try:
            yield f
        finally:
            f.shutdown()
except:
    pass

class FitsLoader:
    """ Remote parallel fits loader, distributes open files to remote
        processes to aid parallization of pyfits operations
    """
    def __init__(self):
        self.fnmap = dict()
        # to synchronize the open file dict
        self.lock = Lock()
        try:
            ncpu = min(mp.cpu_count(), 16)
        except:
            ncpu = 2
        self.procs = []
        self.servers = []
        self.cur = 0
        i = 0
        while len(self.procs) < ncpu:
            port = np.random.randint(5000,20000) + i
            q = mp.Queue()
            t = (port, mp.Process(target=run_server, args=(port,q)))
            import sys
            # work around DFS12172
            if not hasattr(sys.stdout, 'flush'):
                sys.stdout.flush = lambda : None
            if not hasattr(sys.stderr, 'flush'):
                sys.stderr.flush = lambda : None
            t[1].start()
            if q.get():
                self.procs.append(t)
            i += 1
            if i > ncpu + 100:
                raise Exception("Could not start fits loading servers")

        for p in self.procs:
            have_proxy = False
            while not have_proxy:
                try:
                    # ServerProxy not threadsafe in >= 2.7
                    self.servers.append((xmlrpclib.ServerProxy('http://localhost:%d' % p[0],
                                                               allow_none=True), Lock()))
                    have_proxy = True
                except:
                    have_proxy = False

    def _conv_idx(self, idx):
        """ fits extension, None means all extensions"""
        if idx is None:
            return idx
        try:
            return [int(x) for x in idx]
        except:
            return int(idx)

    def files(self):
        """ returns all open files """
        return sorted(self.fnmap.keys())

    def _get_server(self, fn):
        with self.lock:
            if fn in self.fnmap:
                return self.fnmap[fn]
            server, slock = self.servers[self.cur]
            self.fnmap[fn] = server, slock
            self.cur += 1
            if self.cur >= len(self.servers):
                self.cur = 0
            return server, slock

    def open(self, files):
        """ open a file or a list of files """
        if isinstance(files, basestring):
            files = [files]
        res = []
        for fn in files:
            server, slock = self._get_server(fn)
            slock.acquire()
            f = Future(server.pyfits_open, args=(fn,),
                       data=slock, callback=lambda x: x.data.release())
            res.append(f)
        [x.get() for x in res]

    def close(self, fn):
        """ close file """
        self.lock.acquire()
        if fn in self.fnmap:
            server, slock = self.fnmap[fn]
            del self.fnmap[fn]
            self.lock.release()
            with slock:
                server.pyfits_close(fn)

    def getheader(self, fn, idx):
        """ get header from extension idx from file fn """
        idx = self._conv_idx(idx)
        server, slock = self._get_server(fn)
        with slock:
            return server.pyfits_getheader(fn, idx)

    def getdata(self, fn, idx):
        """ get data from extension idx from file fn """
        idx = self._conv_idx(idx)
        server, slock = self._get_server(fn)
        with slock:
            data = server.pyfits_getdata(fn, idx)
        if isinstance(idx, int):
            return deser_array(data)
        if isinstance(idx, list) or idx == None:
            return [deser_array(x) for x in data]

    def getval(self, fn, keyword, idx):
        """ get header value for keyword from extension idx from file fn """
        idx = self._conv_idx(idx)
        server, slock = self._get_server(fn)
        with slock:
            data = server.pyfits_getval(fn, keyword, idx)
        if isinstance(idx, int):
            return data
        if isinstance(idx, list) or idx == None:
            return [deser_array(x) for x in data]

    def getval_all(self, keyword, idx=None, files=None, skip_first=True):
        """ get header value for keyword from extension idx from files fn """
        if files is None:
            files = sorted(self.fnmap.keys())
        idx = self._conv_idx(idx)
        res = []

        for fn in files:
            server, slock = self._get_server(fn)
            slock.acquire()
            res.append(Future(server.pyfits_getval, args=(fn, keyword, idx, skip_first),
                              data=slock, callback=lambda x: x.data.release()))
        if isinstance(idx, int):
            return [x.get() for x in res]
        if isinstance(idx, list) or idx == None:
            return [deser_array(x.get()) for x in res]

    def custom(self, fn, fun, args=None):
        """ execute custom function on fits servers
            function receives opened fits file and args as arguments
        """
        server, slock = self._get_server(fn)
        pfun = pickle.dumps(fun)
        with slock:
            return server.pyfits_custom(fn, pfun, args)

    def len(self, fn):
        """ get number of extensions of a fits file """
        return self.custom(fn, len, None)

    def shutdown(self):
        """ shutdown the fits servers """
        for s, slock in self.servers:
            with slock:
                s.quit()
        del self.servers[:]
        for p in self.procs:
            p[1].join()
        del self.procs[:]


import unittest
try:
    import pyfits
except ImportError:
    import astropy.io.fits as pyfits
import numpy as np
import tempfile
class TestFitsLoader(unittest.TestCase):
    def setUp(self):
        fits = []
        self.next = []
        self.img = []
        fns = [tempfile.NamedTemporaryFile() for x in range(10)]
        fns.append(tempfile.NamedTemporaryFile(prefix=u'БВГДЕЁЖЗИЙК'))
        self.fns = fns
        for i, fn in enumerate(fns):
            hdu = pyfits.HDUList()
            hdu.append(pyfits.PrimaryHDU())
            hdu[0].header.update("TEST", i)
            hdu[0].header.update("TEST2", "bla")
            self.next.append(np.random.randint(1, 100))
            self.img.append([])
            for i in range(self.next[-1] - 1):
                img = pyfits.ImageHDU(data=np.random.uniform(size=(50,50)))
                img.header.update("SUB", i)
                self.img[-1].append(img.data)
                hdu.append(img)
            hdu.writeto(fn.name)
        for fn in fns:
            fits.append(pyfits.open(fn))

    def tearDown(self):
        if hasattr(self, "fl"):
            self.fl.shutdown()

    def test_all(self):
        # should use 2.7 setUpClass and be split into several tests
        fns = self.fns
        fl = FitsLoader()
        self.fl = fl
        for fn in fns:
            fl.open(fn.name)

        self.assertEqual(fl.files(), sorted([x.name for x in fns]))

        for i in range(len(self.fns)):
            header = fl.getheader(fns[i].name, 0)
            self.assertTrue(isinstance(header, dict))
            self.assertTrue("TEST" in header)
            self.assertTrue("TEST2" in header)
            self.assertEqual(header["TEST"], i)
            self.assertEqual(header["TEST2"], "bla")
            next = fl.custom(fns[i].name, len)
            self.assertEqual(next, self.next[i])
            for j in range(1, next):
                self.assertTrue((fl.getdata(fns[i].name, j) == self.img[i][j-1]).all())

        idx = range(1, fl.custom(fns[2].name, len))
        for i, r in enumerate(fl.getdata(fns[2].name, idx)):
            self.assertTrue((r == self.img[2][i]).all())

        self.assertEqual(fl.getval(fns[6].name, "TEST2", 0), "bla")

        self.assertEqual(fl.len(fns[-1].name), self.next[-1])

        self.assertEqual(fl.getval_all("TEST2", idx=0), ["bla"] * len(fns))
        self.assertEqual([list(x) for x in fl.getval_all("SUB", idx=None, files=[fns[0].name, fns[1].name])],
                         [range(self.next[0] - 1), range(self.next[1]-1)])

        f = []
        for fn in fns:
            f.append(Future(fl.getval, args=(fn.name, "TEST2", 0)))
        for x in f:
            self.assertEqual(x.get(), "bla")

        fl.close(fns[6].name)
        for fn in fns[:5]:
            fl.close(fn.name)
        fl.shutdown()

    def test_context(self):
        with get_fitsloader() as fl:
            fl.open([x.name for x in self.fns])

if __name__ == "__main__":
    unittest.main()
